/*
 * AOChat -- library for talking with the Anarchy Online chat servers
 * Copyright (C) 2002  Oskari Saarenmaa <pickett@sumu.org>
 *
 * $Id: keyex.c,v 1.1 2002/11/14 09:53:35 pickett Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

/*
 * based on the PHP class I wrote some time ago.. this C-port is
 * quite unfinished, the only thing it can do at the moment is generate
 * the required login keys (onesided diffie-hellman key exchange with
 * tea encrypted message body)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>
#include <sys/types.h>

#include "keyex.h"
#include "misc.h"

#define STRSIZE 0x1000

void
generate_login_key(char * servkey, char * username, char * password)
{
  mpz_t keya, keyb, keyc, keym, tmp;
  char ckey[STRSIZE], dkey[STRSIZE], str[STRSIZE], pad[9], prefix[9], plain[STRSIZE], crypted[STRSIZE];
  int i, len, nlen;

  mpz_init(keya);
  mpz_init(keyb);
  mpz_init(keyc);
  mpz_init(keym);
  mpz_init(tmp);

  mpz_set_str(keya, "eca2e8c85d863dcdc26a429a71a9815ad052f6139669dd659f98ae159d313d13c6bf2838e10a69b6478b64a24bd054ba8248e8fa778703b418408249440b2c1edd28853e240d8a7e49540b76d120d3b1ad2878b1b99490eb4a2a5e84caa8a91cecbdb1aa7c816e8be343246f80c637abc653b893fd91686cf8d32d6cfe5f2a6f", 16);
  mpz_set_str(keyb, "9c32cc23d559ca90fc31be72df817d0e124769e809f936bc14360ff4bed758f260a0d596584eacbbc2b88bdd410416163e11dbf62173393fbc0c6fefb2d855f1a03dec8e9f105bbad91b3437d8eb73fe2f44159597aa4053cf788d2f9d7012fb8d7c4ce3876f7d6cd5d0c31754f4cd96166708641958de54a6def5657b9f2e92", 16);
  mpz_set_ui(keyc, 5);
  sprintf(str, "%04x%04x%04x%04x%04x%04x%04x%04x", p_random(0xffff), p_random(0xffff), p_random(0xffff), p_random(0xffff), p_random(0xffff), p_random(0xffff), p_random(0xffff), p_random(0xffff));
  mpz_set_str(keym, str, 16);
  
  mpz_powm(tmp, keyc, keym, keya);
  gmp_snprintf(dkey, STRSIZE, "%Zx", tmp);
  mpz_powm(tmp, keyb, keym, keya);
  gmp_snprintf(ckey, STRSIZE, "%Zx", tmp);
  
  snprintf(str, STRSIZE, "%s|%s|%s", username, servkey, password);
  
  if(strlen(ckey) < 32)
  {
    char *dest = ckey + 32 - strlen(ckey);
    memmove(dest, ckey, strlen(ckey));
    memset(ckey, '0', dest-ckey);
  }
  ckey[33] = 0;
  
  for(i=0;i<8;i++)
    prefix[i] = p_random(255);
  prefix[8] = 0;
  
  len  = strlen(str);
  nlen = htonl(len);

  memset(pad, ' ', 8);
  pad[(8-(16+4+len)%8)%8] = 0;

  memcpy(plain, prefix, 8);
  plain[8] = (len>>24)&0xff;
  plain[9] = (len>>16)&0xff;
  plain[10] = (len>>8)&0xff;
  plain[11] = (len>>0)&0xff;
  strcpy(plain+12, str);
  strcpy(plain+12+len, pad);

  aocrypt(crypted, ckey, plain, len+12+strlen(pad));
  
  printf("%s-%s\n", dkey, crypted);
}

void
aocrypt(char *crypted, char *key, char *str, int len)
{
  int i, now[2], pre[2], akey[8], astr[len/4];
  char *p = crypted;
  FILE *foo;
  
  for(i=0; i<4; i++)
    akey[i] = htonl(hexdec(key+i*8, 8));

  for(i=0; i<len/4; i++)
    astr[i] = htonl((str[4*i]<<24) | (str[4*i+1]<<16) | (str[4*i+2]<<8) | (str[4*i+3]));
  
  now[0] = 0; now[1] = 0; pre[0] = 0; pre[1] = 0;

  for(i=0; i<len/4; i+=2)
  {
    now[0] = astr[i]   ^ pre[0];
    now[1] = astr[i+1] ^ pre[1];
    aocrypt_permute(pre, now, akey);
    sprintf(p, "%08x%08x", htonl(pre[0]), htonl(pre[1]));
    p += 16;
  }
}

void
aocrypt_permute(int x[2], int y[2], int key[8])
{
  int a = y[0],
      b = y[1],
      c = 0,
      d = -1640531527,
      i;
  
  for(i=32; i-->0;)
  {
    c += d;
    a += (b << 4 & -16) + key[0] ^ b + c ^ (b >> 5 & 134217727) + key[1];
    b += (a << 4 & -16) + key[2] ^ a + c ^ (a >> 5 & 134217727) + key[3];
  }
  
  x[0] = a;
  x[1] = b;
}
